#!/usr/bin/perl
use Time::Local;
# Quickpatch version 1.0015
# Created by Michael L. Vince / mike@roq.com copyright 2004 ALL RIGHTS RESERVED
#
# Quick and easy source based patch system for FreeBSD.
# Utilizes Perls regular expression power to extract patch commands from official FreeBSD security advisories,
# tested from 4.4 to 5.2
# Get the latest version from http://roq.com/projects/quickpatch/
# Disclaimer: In no way/event shall I be liable for any damages done by this script in any way conceivably possible.

# Requirements, Perl, cvsup setup, wget (/usr/ports/ftp/wget) and a internet connection. Not too hard eh? :)
# Type "./quickpatch.pl updateadv" to download advisories 
# Type "./quickpatch.pl notify" to email some one what would be done in patch mode
# Type "./quickpatch.pl patch" or "./quickpatch.pl patch > big_patch_file" to create patch files
# Type "./quickpatch.pl cvsup" to cvsup to update your source tree
# Optional is pgp (/usr/ports/security/gnupg) to use it type "./quickpatch.pl pgpcheck"

# The release plus security patches branch for FreeBSD that you are following in cvsup. 
# It should always be a long the lines of RELENG_X_X , example RELENG_4_9
$release='RELENG_5_2';
# Ftp server mirror from where to fetch FreeBSD security advisories
$ftpserver="ftp.freebsd.org";
# Path to store patcher program files
$patchdir="/usr/src/";
# Path to store FreeBSD security advisories
$advdir="/var/db/advisories/";
# Path to your kernel rebuild script for source patches that require kernel rebuild
$kernelbuild="/usr/src/buildkernel";
#$kernelbuild="cd /usr/src ; make buildkernel KERNCONF=GENERIC && make installkernel KERNCONF=GENERIC ; reboot";
# Path to your system recompile scipt for patches that require full operating system recompile
$buildworld="/usr/src/buildworld";
#$buildworld="cd /usr/src/ ; make buildworld && make installworld ; reboot";
#Run patch command after creation, for the unloving, default no
$runpatchfile="0";

# Minimum advisory age in hours. This is to make sure you don't patch before your local cvsup server has had a
# chance to recieve the source change update to your branch, in hours
$advisory_age="24";
# Notify email accounts, eg: qw(billg@microsoft.com root@localhost);
@emails = qw(root);

# Hostname of machine, if not set it will extract it get it automatically
#$hostname="mybox";

# CVSUP mirror server, list -> 'grep -i "\* cvsup.*FreeBSD.org" /usr/share/doc/en_US.ISO8859-1/books/handbook/book.txt'
$cvsupserver="cvsup8.FreeBSD.org";

# PGP checking
# If you want to pgp check advisories that use the older FreeBSD Security Officer public key then set this to "0"
# And remember to import the key. These are advisories from FreeBSD-SA-00*.asc to FreeBSD-SA-02*.asc 
# The older PGP public key was stopped being used from FreeBSD 4.7, default yes
$skipoldpgp="1";
#PGP check before patch file creation, highly recommended, default no
$pgpcheck="0";
# If pgpcheck fails send an email, default yes
$pgpemail="1";

################################################################################
# Stuff you probably don't want to change.
# Force recreation of patchfiles that have already been created by this script before. 
$force = "no";
#Gpg binary location (default)
$gpg="/usr/local/bin/gpg";
# Directory to store cvsup configuration file, will get created if it doesn't exist
$supdir="/etc/scripts/";
# CVSUP configuration file which gets recreated everytime cvsup mode is started
$supfile="$supdir"."quickpatch-supfile";
$supfileports="$supdir"."quickpatch-supfile-ports";
#cvsup binary path (default)
$cvsupbin="/usr/local/bin/cvsup";
# Temp adv dir, gets created automatically
$advdirtmp="$advdir"."tmp/";
# Agressive patch mode, instead of matching and only then allowing the patch commands for the most common patch commands 
# that I have seen so far in FreeBSD advisories, this mode blatantly passes through patch commands making it more future
# proof, this should be safe anyway as long as you are PGPing on your advisories and are checking the output of this script 
# before you run it. Other wise it will only allow commands that start with 'cd* or make*' which are the most common patch 
# commands unless its a kernel or base recompile (which you control anyway), default yes
$aggressive="1";
#Wget binary (default)
$wget="/usr/local/bin/wget";
################################################################################
@dirs = ($patchdir,$advdir,$advdirtmp,$supdir);

foreach $dir (@dirs) 	{
if (! -e "$dir")
	{
	print "# Directory $dir does not exist, creating\n";
	mkdir("$dir",0755);
	}
			}

#Sup-file configuration
$sup="*default host=$cvsupserver
*default base=/usr
*default prefix=/usr
*default release=cvs tag=$release
*default delete use-rel-suffix compress
src-all
";

$ports_sup="*default host=$cvsupserver
*default base=/usr
*default prefix=/usr
*default release=cvs tag=.
*default delete use-rel-suffix compress
ports-all
";

if (!$hostname) { $hostname = `/bin/hostname`; chomp($hostname); }

sub getadvlist {
opendir DIR, "$advdir" or die "no dir: $!";
@files = grep /^.*?\.asc$/, readdir DIR;
closedir DIR;
}

sub currentdate
{
($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
($day, $month, $year) = (localtime)[3,4,5];
$year = $year + 1900; $month = $month +1;
$date =  "$day/$month/$year $hour:$min:$sec";
return $date;
}

$date = currentdate();

if ($ARGV[0] eq 'patch') 	{
getadvlist();
foreach $i (@files)	{ find(); }
print "#!/bin/sh\n";
#$mode="patch";

foreach $i (@patchlist)
		{
print "#########################################################################################\n";
print "####### $i\n";
#print "####### $hostname - $date\n";
	if (checkage("$i")) { next; }
	if ($force eq "yes") { patches($i); next; }
	if (oldpatch($i))  { next; } patches($i);
 	if ($runpatchfile) {
	$patchfile =~ s/\.asc//g;
	chomp($patchfile);
	$pfc = "${patchdir}${patchfile}";
	system("$pfc");
	}
	 		}
}
############################################
elsif ($ARGV[0] eq 'notify') 	{
getadvlist();
foreach $i (@files)	{ find(); }
$mode="notify";
foreach $i (@patchlist) {
print "#########################################################################################\n";
print "####### $i\n";
#print "####### $hostname - $date\n";
	if (checkage("$i")) { next; }
if (oldpatch($i)) { next; }
patches($i);
email_patch_check();
	 		} exit;
				}
elsif ($ARGV[0] eq 'pgpcheck')
		{
getadvlist();
foreach $i (@files)	{
gpgcheck($i);
			}
		}

elsif ($ARGV[0] eq 'updateadv') {
if (! -e "$wget")  { print "wget not installed, cd /usr/ports/ftp/wget ; make install\n"; exit; }
system("$wget -nc \"ftp://$ftpserver/pub/FreeBSD/CERT/advisories/FreeBSD-SA*:*.asc\" -P $advdir");
print "Finished\n";
	}
elsif ($ARGV[0] eq 'cvsup')
{
cvsup("$sup","$supfile");
}
elsif ($ARGV[0] eq 'cvsupports')
{
cvsup("$ports_sup","$supfileports");
}
else {
print 'Quickpatch - Easy source based security update system
"./quickpatch.pl updateadv" to download / update advisories db
"./quickpatch.pl patch" or "./quickpatch.pl patch > big_patch_file" to create patch files
"./quickpatch.pl notify" does not do anything but email you commands of what it would do
"./quickpatch.pl cvsup" to cvsup to update your source tree
"./quickpatch.pl cvsupports" to cvsup update your ports tree, default tag is current
"./quickpatch.pl pgpcheck" to PGP check advisories
';
}

sub oldpatch 
{
#$pfc = "$_[0]";
$patchfile = "$_[0]";
$patchfile =~ s/\.asc//g;
chomp($patchfile);
$pfc = "${patchdir}${patchfile}";
if (-e "$pfc") 	{
print "### Skipping $pfc patch file already created \n"; return true;
		} else { return 0; }
}

sub cvsup {
$sdata = "$_[0]"; $sfile = "$_[1]";
open(SUPFILE,">$sfile") || die "Sorry, I couldn't create $supfile\n";
print SUPFILE $sdata;
close(SUPFILE);
if (! -e "$cvsupbin") { print "cvsup not installed, cd /usr/ports/net/cvsup-without-gui ; make install\n"; exit; }
system("$cvsupbin -g -L 2 $sfile");
}

sub find
{
$advfile="${advdir}$i";
open (HT,"$advfile");
while (<HT>)
	{
if (m/UTC.*?$release.*?/si)
    		{
	$data = $_;
	$patch="true";  #print "$advfile \n";
@patchlist = (@patchlist,"$i");
	close(HT); 
		}
	} close(HT);
}

sub gpgcheck 	{
if ($skipoldpgp) {
if ($i =~ /FreeBSD-SA-0[012].*.asc/) { return; }
		 }
		 if (! -e "$gpg") { print "GPG not installed, cd /usr/ports/security/gnupg ; make install\n"; exit; } 
			$gpgout = `$gpg --no-secmem-warning --verify $advdir$_[0] 2>&1`;
#print "$gpgout \n";
if ($gpgout =~ /\ngpg:\s+Good signature from "FreeBSD Security Officer <security-officer\@FreeBSD\.org>"\n/)
{ print "####### PGP check Good: $i \n"; print PF "####### PGP check Good: $i \n";    } else 
	{
print "#############################\nPGP check FAILED: $i be afraid!\n#############################\n"; 
if ($pgpemail) { pgpemail(); }
exit; 
	}
		}

sub pgpemail {
foreach $email (@emails) 	
		{ print "## Sending email to $email \n";
		`echo "PGP check failed! Be afraid! $i" | /usr/bin/mail -s "QuickPatch PGP check failed! $i" $email`;
		print "## Emailed $email \n";  
		}
}

sub checkage() {
$/ ='\n';
open (AG,"${advdir}$_[0]");
while (<AG>)
{
if (/\s+(\d\d\d\d)-(\d\d)-(\d\d)\s(\d\d):(\d\d):(\d\d)\s+UTC\s.*?$release/ || /\s+(\d\d\d\d)\/(\d\d)\/(\d\d)\s(\d\d):(\d\d):(\d\d)\s+UTC\s.*?$release/)
 	{
$datecorrected = "$1-$2-$3 $4:$5:$6"; 
$year = $1; $month = $2; $day = $3; $hours = $4; $minutes = $5; $seconds = $6;
$dce = timegm($seconds, $minutes, $hours, $day, $month-1, $year-1900);
$timenow = time;
$difference = $timenow - $dce;
$seconds    =  $difference % 60;
$difference = ($difference - $seconds) / 60;
$minutes    =  $difference % 60;
$difference = ($difference - $minutes) / 60;
$hours      =  $difference % 24;
#print "$timenow - $dce \n";
if ($difference < $advisory_age) { print "### Advisory too new, minimum age $advisory_age hours was $difference \n";  return true; } else { close(AG); return 0; }
#print "####### Date Corrected: $datecorrected\n####### Hours past since corrected: $difference\n"; 
#print PF "####### Date Corrected: $datecorrected\n####### Hours past since corrected: $difference\n"
 	}
} close(AG); print "No Date corrected found! ${advdir}$_[0] \n"; return true;
		}

sub email_patch_check 	
{
if ($mode) {
	foreach $email (@emails) 	
		{
		`cat "$pf" | /usr/bin/mail -s "QuickPatch notify $topic ${patchfile}" $email`;
		print "## Emailed $email \n";  	
		} unlink $pf;
	 } 
}


sub patches {
$/ = '';
#$/ = '^# cd.*?$';

open (HT,"${advdir}$_[0]");
$patchfile = "$_[0]";
$advfile = "$patchfile";
$patchfile =~ s/\.asc//g;
chomp($patchfile);
$pfc = "${patchdir}${patchfile}";
if ($mode) { $pf = "${advdirtmp}${patchfile}"; } else {
$pf = "${patchdir}${patchfile}"; }
#if (oldpatch()) { return; }

print "####### Stored in file $pf\n";
open(PF,">$pf") || die "Sorry, can't create $pf\n";
#print "$pf\n"; 
chmod 0755, "$pf";
print PF "#!/bin/sh\n";
print PF "####### $i ########################################\n";
#print "####### $i ########################################\n";
if ($pgpcheck) 	{ gpgcheck($_[0]); }

while (<HT>) {
if (m/Topic:\s+/si)
{  $topic = "$_";
chomp($topic); $topic =~ s/\s+\[/ \[/g; 
$topic =~ s/\n//g; $topic =~ s/\s+/ /g; 
print "####### $topic\n"; print PF "####### $topic\n"; 
print "####### Hostname: $hostname - $date\n"; print PF "####### Hostname: $hostname - $date\n";
print "####### Date Corrected: $datecorrected\n####### Hours past since corrected: $difference\n"; 
print PF "####### Date Corrected: $datecorrected\n####### Hours past since corrected: $difference\n"; 
}
#checkage($advfile);

if (m/\w\)\sRe\w+\syour\skernel.*?as\sdescribed\sin/si)
{
# print "####### PATCH CODE Kernel Recompile\n";
 $data = $_;
$data =~ s/\n/ /g;
 print PF "### $data\n";
print "### $data\n";
print "$kernelbuild\n";
print PF "$kernelbuild\n";
close(HT); close(PF);
return;
}

if (m/^Re\w+\syour\skernel\w?\sas\sdescribed\sin/si)
{
# print "####### PATCH CODE Kernel Recompile\n";
 $data = $_;
$data =~ s/\n/ /g;
 print PF "### $data\n";
print "### $data\n";
print "$kernelbuild\n";
print PF "$kernelbuild\n";
close(HT); close(PF);
return;
}

if (m/c\)\sRecompile\s\w+\soperating system\w?\sas\sdescribed\sin/si)
{
 $data = $_;
$data =~ s/\n/ /g;
 print PF "### $data\n";
 print "### $data\n";
print PF "$buildworld\n";
print "$buildworld\n";
close(HT); close(PF);
return;
}

if (m/#\scd/si)
 {
 print "####### Patch Commands \n";  print PF "####### Patch Commands \n";
 $data = $_;
 if ($aggressive) 	{
 #print "# Agressive mode\n";
 $data =~ s/#\s(\w+) /$1 /g; 
 $data =~ s/patch\s+</# patch </g;
 $data =~ s/patch\s+\-p\s+</# patch -p </g;
 $data =~ s/^\[/###\[/g;
 			} else {
#print "# Safe mode \n";
 $data =~ s/#\scd /cd /g;
 $data =~ s/#\smake /make /g;
# Rare types of regexps needed to cover some older 4.x advisories
 $data =~ s/#\senv /env /g;
 $data =~ s/#\sumount /umount /g;
 $data =~ s/#\smount /mount /g;
 $data =~ s/#\skldunload /kldunload /g;
 $data =~ s/#\skldload /kldload /g;
 $data =~ s/#\sinstall /install /g;
 $data =~ s/^\[/###\[/g;
 				}
 print "$data";
 print PF "$data";
 }
	    }
close(HT);
close(PF);
}

